/*
 * This file contains the main function and GUI 
 * functions of the application
 */
#include <string.h>
#include <locale.h>
#include <unistd.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <glib.h>
#include <gtk/gtk.h>

#include <hildon-widgets/hildon-program.h>
#include <hildon-widgets/hildon-banner.h>

#include <libosso.h>
#include <gconf/gconf-client.h>

#include "hci_scan.h"
#include "sdp_search.h"

/* D-BUS service definitions */
#define APPLICATION_DBUS_SERVICE "obexput"

struct UICStruct {
    GtkWidget *fileName;
    GtkListStore *storeDevices;
    GtkWidget *treeDevices;
    GtkFileSelection* fileSelector;
    osso_context_t *osso_context;
};
typedef struct UICStruct * UIComponents;

/* Description des colonnes*/
enum
{
   NAME_COLUMN,
   MAC_COLUMN,
   N_COLUMNS
};

/* 
 * Function sends a file to the destination
 * Returns 0 on success -1 on error
 */
int obexput(const char* filename, char* addr);

static void  quit_event( gpointer   data,
                         guint      callback_action,
                         GtkWidget *menu_item	)
{
    UIComponents uic = (UIComponents) data;
    gtk_main_quit ();
    if (uic->osso_context != NULL)
	osso_deinitialize(uic->osso_context);
    g_free(uic);
}

/* another callback */
static gboolean delete_event( GtkWidget *widget, GdkEvent *event, gpointer data )
{
    quit_event(data, 0, NULL);
    return TRUE;
}

/*
 * Function fills up the list of bluetooth devices
 */
static void fillDevList(UIComponents uic, GSList* devices)
{
    GSList* entry; //list entry
    Device* dev; //bluetooth device
    GtkTreeIter iter; //list iterator
    GtkTreeSelection *select; //selector

    if (devices == NULL)
	return;
    gtk_list_store_clear(GTK_LIST_STORE(uic->storeDevices));
    for (entry = devices; entry; entry = entry->next) {
	dev = (Device*)entry->data;
	if (dev == NULL)
	    continue;
	gtk_list_store_append(GTK_LIST_STORE(uic->storeDevices), &iter);
	gtk_list_store_set(GTK_LIST_STORE(uic->storeDevices), &iter, 
			   NAME_COLUMN,
			   g_strdup(dev->name),
			   MAC_COLUMN, 
			   g_strdup(dev->addr), -1);
	g_free(dev);
	dev = NULL;
    }
    g_slist_free(devices);
    gtk_tree_view_columns_autosize(GTK_TREE_VIEW(uic->treeDevices));
    select = gtk_tree_view_get_selection(GTK_TREE_VIEW (uic->treeDevices));    
}

/*
 * Function is a callback that does device scanning
 */ 
static void scan_cb (GtkButton *widget, gpointer data) 
{
    UIComponents uic = (UIComponents)data; //UI components
    GSList* devices; //devices list

    hildon_banner_show_information(
	GTK_WIDGET(gtk_widget_get_toplevel(uic->treeDevices)), 
	NULL, "Scanning for devices");
    while (gtk_events_pending())
	gtk_main_iteration();
    devices = hci_scan();
    fillDevList(uic, devices);
}

static void send_cb (GtkButton *widget, gpointer data) 
{
    UIComponents uic = (UIComponents)data; //UI components
    GtkTreeSelection *select; //selector
    GList *list; //List of selected items
    gchar* addr; //device MAC address
    GtkTreeModel *model;
    GtkTreeIter iter;
    gchar* fileName = g_strdup(gtk_entry_get_text(GTK_ENTRY(uic->fileName)));

    if (!fileName || fileName[0] == 0)
	return;
    select = gtk_tree_view_get_selection(GTK_TREE_VIEW(uic->treeDevices));
    list = gtk_tree_selection_get_selected_rows(select, &model);
    if (list != NULL && 
	gtk_tree_model_get_iter(model, &iter, (GtkTreePath *)list->data)) 
    {
	gtk_tree_model_get(model, &iter, MAC_COLUMN, &addr, -1);
	if (addr != NULL) 
	{
	    printf("Ready to send %s to %s\n", fileName, addr);
	    hildon_banner_show_information(
		GTK_WIDGET(gtk_widget_get_toplevel(uic->treeDevices)), 
		NULL, "Sending file");
	    while (gtk_events_pending())
		gtk_main_iteration();
	    if (obexput(fileName, addr) < 0)
		hildon_banner_show_information(
		    GTK_WIDGET(gtk_widget_get_toplevel(uic->treeDevices)), 
		    NULL, "File transmission failed");
	    else
		hildon_banner_show_information(
		    GTK_WIDGET(gtk_widget_get_toplevel(uic->treeDevices)), 
		    NULL, "File transmission done");
	}
	g_free(addr);
    }
    g_list_free(list);
    g_free(fileName);
}

void browseOk(GtkWidget *widget, gpointer data) 
{
    UIComponents uic = (UIComponents)data; //UI components

    gtk_entry_set_text(GTK_ENTRY(uic->fileName),
		       gtk_file_selection_get_filename(uic->fileSelector));
    gtk_widget_destroy(GTK_WIDGET(uic->fileSelector));
}
/*
 * Function is called when a name is selected
 */
static void browse_cb(GtkTreeSelection *treeselection, gpointer data)
{
    UIComponents uic = (UIComponents)data; //UI components

    uic->fileSelector = GTK_FILE_SELECTION(gtk_file_selection_new ("Files"));
    /* hide all the buttons etc which manipulate file like create, rename etc */
    gtk_file_selection_hide_fileop_buttons(uic->fileSelector);
    gtk_file_selection_set_select_multiple (uic->fileSelector, FALSE);
    g_signal_connect (GTK_FILE_SELECTION (uic->fileSelector)->ok_button,
                      "clicked",
                      G_CALLBACK (browseOk),
                      data);
    g_signal_connect_swapped (GTK_FILE_SELECTION (uic->fileSelector)->ok_button,
			      "clicked",
			      G_CALLBACK (gtk_widget_destroy), 
			      uic->fileSelector);

    g_signal_connect_swapped (GTK_FILE_SELECTION (uic->fileSelector)->cancel_button,
			      "clicked",
			      G_CALLBACK (gtk_widget_destroy),
			      uic->fileSelector); 
    gtk_file_selection_set_filename(uic->fileSelector, 
				    gtk_entry_get_text(GTK_ENTRY(uic->fileName)));
    /* Display that dialog */
    gtk_widget_show(GTK_WIDGET(uic->fileSelector));
}

/*
 * Function reads bluetoot devices registered in the system
 * It reads them from /system/bluetooth/
 * return 0 on success -1 on error 
 */ 
int getSysDevices(UIComponents uic)
{
    int ret = 0; //The return value
    GSList* entry; //list entry
    gchar* btaddr; //pointer to the bluetooth address string
    GConfClient* client = gconf_client_get_default();
    GSList* devices = NULL; //list of bluetooth devices
    if (!client)
	return -1;
    GSList* devaddr = gconf_client_all_dirs(client, "/system/bluetooth/device", NULL);
    gtk_list_store_clear(GTK_LIST_STORE(uic->storeDevices));
    if (!devaddr)
	goto out;
    for (entry = devaddr; entry; entry=entry->next) 
    {
	if (entry->data == NULL)
	    continue;
	btaddr = strrchr((gchar*)entry->data, '/');
	if (btaddr)
	{
	    gchar* newstr = g_strdup(btaddr + 1);
	    g_free(entry->data);
	    entry->data = newstr;
	}
    }
    
    devices = hci_scan_names(devaddr);
    fillDevList(uic, devices);
out:
    for (entry=devaddr; entry; entry=entry->next) 
    {
	if (entry->data == NULL)
	    continue;
	g_free(entry->data);
	entry->data = NULL;
    }
    g_slist_free(devaddr);
    g_object_unref (G_OBJECT(client));
    return ret;
}


int main( int argc, char *argv[] )
{
    /* GtkWidget is the storage type for widgets */
    GtkWidget *button;
    GtkWidget *browseButton;
    GtkCellRenderer *renderer;
    GtkTreeViewColumn *column;
    GtkTreeSelection *select;
    GtkWidget *table;
    GtkWidget *scrolled_window;
	
    HildonProgram *app;
    HildonWindow *appview;
	
    /* Initialize OSSO and install D-Bus handler */
    UIComponents uic = (UIComponents) g_malloc (sizeof(struct UICStruct));

    /* This is called in all GTK applications. Arguments are parsed
     * from the command line and are returned to the application. */
    gtk_init (&argc, &argv);

    /* Create a new window */
    app = HILDON_PROGRAM( hildon_program_get_instance() );
    g_set_application_name("osso-obexput");
	
    /* Create HildonAppView and set it to HildonApp */
    appview = HILDON_WINDOW(hildon_window_new());
    hildon_program_add_window( app, appview );


    uic->osso_context = osso_initialize(APPLICATION_DBUS_SERVICE, "0.0.1", TRUE, NULL);

    /* Check that initialization was ok */
    if (uic->osso_context == NULL) {
	hildon_banner_show_information(GTK_WIDGET(appview), NULL, 
				       "osso_initialize error\n");
	printf("osso_initialize error\n");
	return OSSO_ERROR;
    }

    /* Here we just set a handler for delete_event that immediately
     *      * exits GTK. */
    g_signal_connect (G_OBJECT (appview), "delete_event",
		      G_CALLBACK (delete_event), (gpointer) uic);

    //Create the whole table
    table = gtk_table_new(9, 5, TRUE);
    gtk_container_add (GTK_CONTAINER (appview), table);
    gtk_table_set_col_spacings(GTK_TABLE(table), 1);
	
    /* Add a field & 1 button */
    uic->fileName = gtk_entry_new();
    gtk_table_attach_defaults( GTK_TABLE(table), uic->fileName, 0, 4, 0, 1);

    /* Buttons */
    browseButton = gtk_button_new_with_label ("Browse");
    GTK_WIDGET_SET_FLAGS(browseButton, GTK_CAN_DEFAULT);
    g_signal_connect (G_OBJECT (browseButton), "clicked",
		      G_CALLBACK (browse_cb), (gpointer) uic);
    gtk_table_attach_defaults(GTK_TABLE(table), browseButton, 4, 5, 0, 1);
    gtk_widget_grab_default(browseButton);

    // Devices entry
    uic->storeDevices = gtk_list_store_new (N_COLUMNS, G_TYPE_STRING, G_TYPE_STRING);

    scrolled_window = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled_window), 
				   GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	
    uic->treeDevices = gtk_tree_view_new_with_model(GTK_TREE_MODEL(uic->storeDevices));
	
    GTK_WIDGET_UNSET_FLAGS(uic->treeDevices, GTK_CAN_FOCUS);
    gtk_scrolled_window_set_shadow_type(scrolled_window, GTK_SHADOW_ETCHED_IN);
    gtk_container_add(GTK_CONTAINER(scrolled_window), uic->treeDevices);
    gtk_table_attach_defaults(GTK_TABLE(table), scrolled_window, 0, 3, 1, 9);

    //Set names
    renderer = gtk_cell_renderer_text_new ();
    column = gtk_tree_view_column_new_with_attributes ("Name", renderer, "text", 
						       NAME_COLUMN, NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (uic->treeDevices), column);

    renderer = gtk_cell_renderer_text_new ();
    column = gtk_tree_view_column_new_with_attributes ("MAC", renderer, "text", 
						       MAC_COLUMN, NULL);
    gtk_tree_view_append_column (GTK_TREE_VIEW (uic->treeDevices), column);
    gtk_tree_view_column_set_visible (column, FALSE);
    gtk_tree_view_set_search_column(GTK_TREE_VIEW (uic->treeDevices), NAME_COLUMN);
    gtk_tree_sortable_set_sort_column_id(GTK_TREE_SORTABLE(
					     GTK_LIST_STORE(uic->storeDevices)), 
					 NAME_COLUMN, GTK_SORT_ASCENDING);
    /* Setup the selection handler */
    select = gtk_tree_view_get_selection (GTK_TREE_VIEW (uic->treeDevices));
    gtk_tree_selection_set_mode (select, GTK_SELECTION_SINGLE);
	
    //Set buttons
    button = gtk_button_new_with_label ("Scan");
    g_signal_connect (G_OBJECT (button), "clicked",
		      G_CALLBACK (scan_cb), (gpointer) uic);
    gtk_table_attach_defaults(GTK_TABLE(table), button, 4, 5, 2, 3);
	
    button = gtk_button_new_with_label ("Send");
    g_signal_connect (G_OBJECT (button), "clicked",
		      G_CALLBACK (send_cb), (gpointer) uic);
    gtk_table_attach_defaults(GTK_TABLE(table), button, 4, 5, 1, 2);
	
    gtk_widget_show_all (GTK_WIDGET(appview));

    //Fill BT devices from the system config
    getSysDevices(uic);	

    /* Rest in gtk_main and wait for the fun to begin! */
    gtk_main ();

    return OSSO_OK;
}

